/* Native-dependent code for GNU/Linux x86-64.

   Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006
   Free Software Foundation, Inc.
   Contributed by Jiri Smid, SuSE Labs.

   This file is part of GDB.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor,
   Boston, MA 02110-1301, USA.  */

#include "defs.h"
#include "inferior.h"
#include "gdbcore.h"
#include "regcache.h"
#include "linux-nat.h"
#include "amd64-linux-tdep.h"

#include "gdb_assert.h"
#include "gdb_string.h"
#include <sys/ptrace.h>
#include <sys/debugreg.h>
#include <sys/syscall.h>
#include <sys/procfs.h>
#include <asm/prctl.h>
/* FIXME ezannoni-2003-07-09: we need <sys/reg.h> to be included after
   <asm/ptrace.h> because the latter redefines FS and GS for no apparent
   reason, and those definitions don't match the ones that libpthread_db
   uses, which come from <sys/reg.h>.  */
/* ezannoni-2003-07-09: I think this is fixed. The extraneous defs have
   been removed from ptrace.h in the kernel.  However, better safe than
   sorry.  */
#include <asm/ptrace.h>
#include <sys/reg.h>
#include "gdb_proc_service.h"

/* Prototypes for supply_gregset etc.  */
#include "gregset.h"

#include "amd64-tdep.h"
#include "i386-linux-tdep.h"
#include "amd64-nat.h"

/* Mapping between the general-purpose registers in GNU/Linux x86-64
   `struct user' format and GDB's register cache layout.  */

static int amd64_linux_gregset64_reg_offset[] =
{
  RAX * 8, RBX * 8,		/* %rax, %rbx */
  RCX * 8, RDX * 8,		/* %rcx, %rdx */
  RSI * 8, RDI * 8,		/* %rsi, %rdi */
  RBP * 8, RSP * 8,		/* %rbp, %rsp */
  R8 * 8, R9 * 8,		/* %r8 ... */
  R10 * 8, R11 * 8,
  R12 * 8, R13 * 8,
  R14 * 8, R15 * 8,		/* ... %r15 */
  RIP * 8, EFLAGS * 8,		/* %rip, %eflags */
  CS * 8, SS * 8,		/* %cs, %ss */
  DS * 8, ES * 8,		/* %ds, %es */
  FS * 8, GS * 8,		/* %fs, %gs */
  -1, -1, -1, -1, -1, -1, -1, -1,
  -1, -1, -1, -1, -1, -1, -1, -1,
  -1, -1, -1, -1, -1, -1, -1, -1,
  -1, -1, -1, -1, -1, -1, -1, -1, -1,
  ORIG_RAX * 8
};


/* Mapping between the general-purpose registers in GNU/Linux x86-64
   `struct user' format and GDB's register cache layout for GNU/Linux
   i386.

   Note that most GNU/Linux x86-64 registers are 64-bit, while the
   GNU/Linux i386 registers are all 32-bit, but since we're
   little-endian we get away with that.  */

/* From <sys/reg.h> on GNU/Linux i386.  */
static int amd64_linux_gregset32_reg_offset[] =
{
  RAX * 8, RCX * 8,		/* %eax, %ecx */
  RDX * 8, RBX * 8,		/* %edx, %ebx */
  RSP * 8, RBP * 8,		/* %esp, %ebp */
  RSI * 8, RDI * 8,		/* %esi, %edi */
  RIP * 8, EFLAGS * 8,		/* %eip, %eflags */
  CS * 8, SS * 8,		/* %cs, %ss */
  DS * 8, ES * 8,		/* %ds, %es */
  FS * 8, GS * 8,		/* %fs, %gs */
  -1, -1, -1, -1, -1, -1, -1, -1,
  -1, -1, -1, -1, -1, -1, -1, -1,
  -1, -1, -1, -1, -1, -1, -1, -1, -1,
  ORIG_RAX * 8			/* "orig_eax" */
};


/* Transfering the general-purpose registers between GDB, inferiors
   and core files.  */

/* Fill GDB's register cache with the general-purpose register values
   in *GREGSETP.  */

void
supply_gregset (elf_gregset_t *gregsetp)
{
  amd64_supply_native_gregset (current_regcache, gregsetp, -1);
}

/* Fill register REGNUM (if it is a general-purpose register) in
   *GREGSETP with the value in GDB's register cache.  If REGNUM is -1,
   do this for all registers.  */

void
fill_gregset (elf_gregset_t *gregsetp, int regnum)
{
  amd64_collect_native_gregset (current_regcache, gregsetp, regnum);
}

/* Transfering floating-point registers between GDB, inferiors and cores.  */

/* Fill GDB's register cache with the floating-point and SSE register
   values in *FPREGSETP.  */

void
supply_fpregset (elf_fpregset_t *fpregsetp)
{
  amd64_supply_fxsave (current_regcache, -1, fpregsetp);
}

/* Fill register REGNUM (if it is a floating-point or SSE register) in
   *FPREGSETP with the value in GDB's register cache.  If REGNUM is
   -1, do this for all registers.  */

void
fill_fpregset (elf_fpregset_t *fpregsetp, int regnum)
{
  amd64_collect_fxsave (current_regcache, regnum, fpregsetp);
}


/* Transferring arbitrary registers between GDB and inferior.  */

/* Fetch register REGNUM from the child process.  If REGNUM is -1, do
   this for all registers (including the floating point and SSE
   registers).  */

static void
amd64_linux_fetch_inferior_registers (int regnum)
{
  int tid;

  /* GNU/Linux LWP ID's are process ID's.  */
  tid = TIDGET (inferior_ptid);
  if (tid == 0)
    tid = PIDGET (inferior_ptid); /* Not a threaded program.  */

  if (regnum == -1 || amd64_native_gregset_supplies_p (regnum))
    {
      elf_gregset_t regs;

      if (ptrace (PTRACE_GETREGS, tid, 0, (long) &regs) < 0)
	perror_with_name (_("Couldn't get registers"));

      amd64_supply_native_gregset (current_regcache, &regs, -1);
      if (regnum != -1)
	return;
    }

  if (regnum == -1 || !amd64_native_gregset_supplies_p (regnum))
    {
      elf_fpregset_t fpregs;

      if (ptrace (PTRACE_GETFPREGS, tid, 0, (long) &fpregs) < 0)
	perror_with_name (_("Couldn't get floating point status"));

      amd64_supply_fxsave (current_regcache, -1, &fpregs);
    }
}

/* Store register REGNUM back into the child process.  If REGNUM is
   -1, do this for all registers (including the floating-point and SSE
   registers).  */

static void
amd64_linux_store_inferior_registers (int regnum)
{
  int tid;

  /* GNU/Linux LWP ID's are process ID's.  */
  tid = TIDGET (inferior_ptid);
  if (tid == 0)
    tid = PIDGET (inferior_ptid); /* Not a threaded program.  */

  if (regnum == -1 || amd64_native_gregset_supplies_p (regnum))
    {
      elf_gregset_t regs;

      if (ptrace (PTRACE_GETREGS, tid, 0, (long) &regs) < 0)
	perror_with_name (_("Couldn't get registers"));

      amd64_collect_native_gregset (current_regcache, &regs, regnum);

      if (ptrace (PTRACE_SETREGS, tid, 0, (long) &regs) < 0)
	perror_with_name (_("Couldn't write registers"));

      if (regnum != -1)
	return;
    }

  if (regnum == -1 || !amd64_native_gregset_supplies_p (regnum))
    {
      elf_fpregset_t fpregs;

      if (ptrace (PTRACE_GETFPREGS, tid, 0, (long) &fpregs) < 0)
	perror_with_name (_("Couldn't get floating point status"));

      amd64_collect_fxsave (current_regcache, regnum, &fpregs);

      if (ptrace (PTRACE_SETFPREGS, tid, 0, (long) &fpregs) < 0)
	perror_with_name (_("Couldn't write floating point status"));

      return;
    }
}


static unsigned long
amd64_linux_dr_get (int regnum)
{
  int tid;
  unsigned long value;

  /* FIXME: kettenis/2001-01-29: It's not clear what we should do with
     multi-threaded processes here.  For now, pretend there is just
     one thread.  */
  tid = PIDGET (inferior_ptid);

  /* FIXME: kettenis/2001-03-27: Calling perror_with_name if the
     ptrace call fails breaks debugging remote targets.  The correct
     way to fix this is to add the hardware breakpoint and watchpoint
     stuff to the target vectore.  For now, just return zero if the
     ptrace call fails.  */
  errno = 0;
  value = ptrace (PT_READ_U, tid,
		  offsetof (struct user, u_debugreg[regnum]), 0);
  if (errno != 0)
#if 0
    perror_with_name (_("Couldn't read debug register"));
#else
    return 0;
#endif

  return value;
}

static void
amd64_linux_dr_set (int regnum, unsigned long value)
{
  int tid;

  /* FIXME: kettenis/2001-01-29: It's not clear what we should do with
     multi-threaded processes here.  For now, pretend there is just
     one thread.  */
  tid = PIDGET (inferior_ptid);

  errno = 0;
  ptrace (PT_WRITE_U, tid, offsetof (struct user, u_debugreg[regnum]), value);
  if (errno != 0)
    perror_with_name (_("Couldn't write debug register"));
}

void
amd64_linux_dr_set_control (unsigned long control)
{
  amd64_linux_dr_set (DR_CONTROL, control);
}

void
amd64_linux_dr_set_addr (int regnum, CORE_ADDR addr)
{
  gdb_assert (regnum >= 0 && regnum <= DR_LASTADDR - DR_FIRSTADDR);

  amd64_linux_dr_set (DR_FIRSTADDR + regnum, addr);
}

void
amd64_linux_dr_reset_addr (int regnum)
{
  gdb_assert (regnum >= 0 && regnum <= DR_LASTADDR - DR_FIRSTADDR);

  amd64_linux_dr_set (DR_FIRSTADDR + regnum, 0L);
}

unsigned long
amd64_linux_dr_get_status (void)
{
  return amd64_linux_dr_get (DR_STATUS);
}


/* This function is called by libthread_db as part of its handling of
   a request for a thread's local storage address.  */

ps_err_e
ps_get_thread_area (const struct ps_prochandle *ph,
                    lwpid_t lwpid, int idx, void **base)
{
  if (gdbarch_ptr_bit (current_gdbarch) == 32)
    {
      /* The full structure is found in <asm-i386/ldt.h>.  The second
	 integer is the LDT's base_address and that is used to locate
	 the thread's local storage.  See i386-linux-nat.c more
	 info.  */
      unsigned int desc[4];

      /* This code assumes that "int" is 32 bits and that
	 GET_THREAD_AREA returns no more than 4 int values.  */
      gdb_assert (sizeof (int) == 4);	
#ifndef PTRACE_GET_THREAD_AREA
#define PTRACE_GET_THREAD_AREA 25
#endif
      if  (ptrace (PTRACE_GET_THREAD_AREA, 
		   lwpid, (void *) (long) idx, (unsigned long) &desc) < 0)
	return PS_ERR;
      
      /* Extend the value to 64 bits.  Here it's assumed that a "long"
	 and a "void *" are the same.  */
      (*base) = (void *) (long) desc[1];
      return PS_OK;
    }
  else
    {
      /* This definition comes from prctl.h, but some kernels may not
         have it.  */
#ifndef PTRACE_ARCH_PRCTL
#define PTRACE_ARCH_PRCTL      30
#endif
      /* FIXME: ezannoni-2003-07-09 see comment above about include
	 file order.  We could be getting bogus values for these two.  */
      gdb_assert (FS < ELF_NGREG);
      gdb_assert (GS < ELF_NGREG);
      switch (idx)
	{
	case FS:
	  if (ptrace (PTRACE_ARCH_PRCTL, lwpid, base, ARCH_GET_FS) == 0)
	    return PS_OK;
	  break;
	case GS:
	  if (ptrace (PTRACE_ARCH_PRCTL, lwpid, base, ARCH_GET_GS) == 0)
	    return PS_OK;
	  break;
	default:                   /* Should not happen.  */
	  return PS_BADADDR;
	}
    }
  return PS_ERR;               /* ptrace failed.  */
}


static void (*super_post_startup_inferior) (ptid_t ptid);

static void
amd64_linux_child_post_startup_inferior (ptid_t ptid)
{
  i386_cleanup_dregs ();
  super_post_startup_inferior (ptid);
}


/* ------------------------------------------------------------------------- */

#include "nvGpuDebugServer.h"
#include "nvPtxStorage.h"
#include "stdRangeMap.h"

gpudbgGpuCoordinates previous_gpu_thread;

static stdRangeMap_t ptxAddresses;
static CORE_ADDR     saved_raw_pc;
static int           saved_raw_pc_valid;

static void    (*saved_to_kill) (void);
static void    (*saved_to_mourn_inferior) (void);
static ptid_t  (*saved_to_wait) (ptid_t, struct target_waitstatus *);
static void    (*saved_to_resume) (ptid_t tid, int i, enum target_signal ts);
static void    (*saved_to_create_inferior) (char *, char *, char **, int);
static void    (*saved_to_fetch_registers) (int regno);
static void    (*saved_to_store_registers) (int regno);
static int     (*saved_to_insert_breakpoint) (struct bp_target_info *bp_tgt);
static int     (*saved_to_remove_breakpoint) (struct bp_target_info *bp_tgt);

static LONGEST (*saved_to_xfer_partial) (struct target_ops *ops,
				enum target_object object, const char *annex,
				gdb_byte *readbuf, const gdb_byte *writebuf,
				ULONGEST offset, LONGEST len);


static void wrapper_to_kill(void)
{
    gpudbg_cleanup();
    saved_to_kill();
}


void gdbg_flush_saved_thread_state()
{
    if (saved_raw_pc_valid) { 
        regcache_raw_write(current_regcache,PC_REGNUM,&saved_raw_pc);
        saved_raw_pc_valid= 0;
    }
}


static void wrapper_to_mourn_inferior(void)
{
    if (ptxAddresses) {
        gpudbg->rangemapDelete(ptxAddresses);
        ptxAddresses= NULL;
    }
    
    saved_raw_pc_valid= 0;
    gpudbg_cleanup();
    saved_to_mourn_inferior();
}


static void wrapper_to_resume (ptid_t tid, int sstep, enum target_signal ts)
{
    gpudbgGPUContext gpuContext;
    int gpuContextTid;

    gdbg_flush_saved_thread_state();

    gpuContextTid = tid.pid == -1 
        ? inferior_ptid.lwp ? inferior_ptid.lwp : inferior_ptid.pid
        : tid.lwp ? tid.lwp : tid.pid;
    
    /*
     * See patch at infrun.c, function currently_stepping (Return Code is 2):
     */
    if (sstep != 2) {
       /*
        * From gnu-nat.c, line 1924:
        *       TID  STEP:
        *       -1   true   Single step the current thread allowing other threads to run.
        *       -1   false  Continue the current thread allowing other threads to run.
        *       X    true   Single step the given thread, don't allow any others to run.
        *       X    false  Continue the given thread, do not allow any others to run.
        *       (Where X, of course, is anything except -1)
        */

        gpudbg->ResumeGpus(gpuContextTid, sstep, tid.pid == -1 ? True : False);
    }

    /* Now we determine how to resume the host.  We always need the host to run in
     * the case of a GPU context, otherwise we cannot detect GPU termination. */
    if (gpudbg->GetGPUContext(gpuContextTid, &gpuContext)) {
        /* We have a context, so let the host go, we'll send the trap for the GPU */
        saved_to_resume(tid,0,ts);

        /* If we single stepped the GPU, send the trap here.
         * The target layer handles this for breakpoints. */
        if (sstep) {
            kill(gpuContextTid, SIGTRAP);
        }
    }
    else {
        /* Always resume the host if there is no gpu context */
        saved_to_resume(tid,sstep,ts);
    }

}

static ptid_t wrapper_to_wait (ptid_t tid, struct target_waitstatus *ws)
{
    gpudbgGPUContext gpuContext;
    ptid_t r= saved_to_wait(tid,ws);

    /* We need to make sure that saved_to_wait hasn't neglected the GPU.
     * It is possible that the lwp field inside of r (above) has been changed
     * to a different lwp in the host process. We check to see that we had a
     * gpu thread in our previous context, then we also check to see that we
     * still have an active GPU context. If so, we set the return value of
     * this function accordingly. */

    /* Check to see if we have a gpu context with the current r.lwp.
     * We only update if we do not have a gpu context */
    if (ptid_get_lwp(r) != 0 && !gpudbg->GetGPUContext(r.lwp, &gpuContext)){
        if (previous_gpu_thread && ptid_get_lwp(inferior_ptid) != 0){
            /* We know that we had a previous gpu thread and that there is
             * a specified lwp field in inferior_ptid */
            if (gpudbg->GetGPUContext(inferior_ptid.lwp, &gpuContext)){
                /* We know that the lwp field matches that registered by the client */
                r = inferior_ptid;
            }
        }
    }

   /*
    * Most cleanly, gpudbg_initialize() should have been called
    * at inferior creation. It is done lazily here because initialization
    * attaches to special debugger symbols in the cuda library, and this 
    * library is loaded later during executable startup. We therefore
    * simply wait until we see it appear. 
    * It is guaranteed that we eventually catch it, because thread creation 
    * and dll loading will trigger hidden debugger breakpoints.
    */
    gpudbg_initialize();
    gpudbg->SuspendGpus();
    
   /*
    * The clean way of getting to know all host addresses that 
    * correspond with device function code is to place an observer
    * in the shared library module, to get notified whenever a new
    * shared library gets loaded or unloaded. However, this is another
    * intrusion in the gdb code, so we do it lazily: we rescan the 
    * list of currently loaded libraries each time we get control back:
    */
    if (ptxAddresses) {
        gpudbg->rangemapDelete(ptxAddresses);
        ptxAddresses= NULL;
    }

    return r;
}

static void wrapper_to_create_inferior (char *a, char *b, char **c, int d)
{
    saved_to_create_inferior(a,b,c,d);
    gpudbg_initialize();
}

static LONGEST wrapper_to_xfer_partial(struct target_ops *ops,
				enum target_object object, const char *annex,
				gdb_byte *readbuf, const gdb_byte *writebuf,
				ULONGEST offset, LONGEST len)
{
    LONGEST result= 0;
    
    if (object == TARGET_OBJECT_MEMORY) {
        if (readbuf) {
            result= gpudbg->ReadMemory(offset,readbuf,len);
        } else
        if (writebuf) {
            result= gpudbg->WriteMemory(offset,writebuf,len);
        }
    }
    
    if (!result) {
        result= saved_to_xfer_partial(ops,object,annex,readbuf,writebuf,offset,len);
    }
            
    return result;
}

static void wrapper_to_fetch_registers (int regno)
{
    saved_to_fetch_registers (regno);
}

static void wrapper_to_store_registers (int regno)
{
    saved_to_store_registers (regno);
}

static int wrapper_to_insert_breakpoint (struct bp_target_info *bp_tgt)
{
  if (gpudbg->SetBreakpoint(bp_tgt->placed_address,bp_tgt->owner->owner->gpu_thread,1)) {
    return 0;
  } else {
    return saved_to_insert_breakpoint(bp_tgt);
  }
}

static int wrapper_to_remove_breakpoint (struct bp_target_info *bp_tgt)
{
  if (gpudbg->SetBreakpoint(bp_tgt->placed_address,bp_tgt->owner->owner->gpu_thread,0)) {
    return 0;
  } else {
    return saved_to_remove_breakpoint(bp_tgt);
  }
}


/* ------------------------------------------------------------------------- */

/*
 * The full_massage parameter having value False causes gpudbg_massage_call_chain 
 * to only patch the topmost stack frame (describing the current pc).
 * This is used when it is known that no stack frame up/down traversal
 * is necessary (see info gpu threads command).
 * A better solution would be to let full massaging be called lazily by
 * function get_prev_frame:
 */

void gpudbg_massage_call_chain( ptid_t ptid, int full_massage )
{
    gpudbgGPUContext gpuContext;
    int tid = TIDGET (ptid);

   /*
    * Inspect the break pc, and if this is a GPU address,
    * then precache the frame_info chain up to the device code
    * 'caller', while kicking out all frames that correspond to the cuda
    * runtime launching functions. This makes it appear in the debugger
    * as if the application directly 'called' the device function.
    */
    if (gpudbg->GetGPUContext(tid, &gpuContext)) { 
        struct frame_info *bottom,*frame;
        int                morePrev = 1;
        CORE_ADDR          low,high,raw_pc;
        char              *name;
        
       /*
        * Restore original pc in host thread, otherwise
        * frame unwinding will use the wrong debugger info:
        */
        bottom = get_current_frame();
        
        gdbg_flush_saved_thread_state();

        raw_pc = read_register_pid (PC_REGNUM, ptid);
        kickout_intermediate_frames(bottom,bottom,ADDR_BITS_REMOVE(raw_pc));
        
       /*
        * While unwinding the stack, search for the 
        * device function stub called by the application,
        * and then kick out all intermediate frames from the
        * frame cache:
        */
        frame = get_prev_frame(bottom);
        
        if ( full_massage
          && find_pc_partial_function (gpuContext.functionWrapper, &name, &low, &high)
           ) {
            while (frame) {
                CORE_ADDR pc = get_frame_pc(frame);
                if (pc >= low && pc < high) { break; }
                frame = get_prev_frame(frame);
            }

            while (frame && morePrev) {
                frame = get_prev_frame(frame);
                morePrev--;
            }
        }

        if (frame) {
            CORE_ADDR pc_val   = gpuContext.pc;
            saved_raw_pc       = raw_pc;
            saved_raw_pc_valid = 1;
            
            get_frame_id(frame); // evaluate the frame descriptor
            kickout_intermediate_frames(frame,bottom,gpuContext.pc);
            regcache_raw_write(current_regcache,PC_REGNUM,&pc_val);
        }
    }
}

/* ------------------------------------------------------------------------- */

#include "filenames.h"
#include "objfiles.h"

/*
 * Read all known source link sections (which consist of pairs of filenames,
 * <host source file,device source file>. If pst->filename appears to be such
 * a host source filename, then also load the symbtab of the corresponding 
 * device source file name. ALthough duplicates are not likely, neither host 
 * source file name nor device source file name is supposed to be unique 
 * (they may be temp file names). If they are not, then we might at most 
 * load too many symbol tables, but this is not a problem for the debugger.
 * (this function is a patch to dwarf2read.c)
 */
void gpudbg_loadDeviceInfo( char *fileName, struct partial_symtab *pst)
{
    struct section_table *secp;
    for (secp = current_target.to_sections;
         secp < current_target.to_sections_end;
         secp++) {
        if (strcmp(secp->the_bfd_section->name,PTX_SOURCELINK_SEGMENT_NAME)==0) {
            int   sectionSize = secp->endaddr - secp->addr; 
            char *sourceLinks = alloca(sectionSize);
            char *sourceBound = sourceLinks + sectionSize;

            target_read_memory(secp->addr,sourceLinks,sectionSize);

            while (sourceLinks < sourceBound) {
                int   size     = strlen(sourceLinks)+1;
                char *hostName = sourceLinks;
                char *devName  = strchr(sourceLinks,',');

                // --- we need some sanity checking here ---

                *(devName++) = 0;
                sourceLinks += size;

                if (FILENAME_CMP (hostName, fileName) == 0) {
                    struct partial_symtab *ps;
                    ALL_OBJFILE_PSYMTABS (pst->objfile, ps)
                    {
                        if (FILENAME_CMP (devName, ps->filename) == 0) {
                            if (!ps->readin)
                                dwarf2_psymtab_to_symtab(ps);
                        }
                    }
                }
                
               /*
                * creep past section padding:
                */
                while ((sourceLinks < sourceBound) && *sourceLinks==0) {
                    sourceLinks++;
                }
            }
        }
    }
}

/* ------------------------------------------------------------------------- */

#include "filenames.h"
#include "objfiles.h"

/*
 * Read all known ptx virtual code sections, and record 
 * the virtual ptx code addresses in a range map:
 */
int gpudbg_is_ptx_address( CORE_ADDR address )
{
    if (!ptxAddresses) {
        struct section_table *secp;
        
        ptxAddresses = gpudbg->rangemapCreate();
        
        for (secp = current_target.to_sections;
             secp < current_target.to_sections_end;
             secp++) {
            if (strcmp(secp->the_bfd_section->name,PTX_NVSEGMENT_NAME(10))==0) {
                rangemapDomain_t dom;
                
                dom.start  = secp->addr;
                dom.length = secp->endaddr - secp->addr; 
                
                gpudbg->rangemapDefine( ptxAddresses,  dom, (Pointer)(Address)True );
            }
        }
    }
    
    return gpudbg->rangemapApply( ptxAddresses, address ) != NULL;
}

/* ------------------------------------------------------------------------- */

#include "ui-out.h"
#include "command.h"
#include "gdb.h"
#include "exceptions.h"
#include "gdbcmd.h"

String GetGpuCoordsRepresentation(Bool isMask, gpudbgGpuCoordinates mask)
{
    Int    defaultIndex= isMask ? -1 : 0;
    Int    blockSpec = 0;

    static Char  result[1000];
           Char *p= result;
    
    if (!isMask || mask->bX != defaultIndex || mask->bY != defaultIndex) {
        blockSpec = 1;

        p += sprintf(p, "<<<(");
        
        if (!isMask || mask->bX != defaultIndex) { 
            p += sprintf(p, "%d",mask->bX ); 
        }
        
        if (mask->bY != defaultIndex) { 
            p += sprintf(p, ",%d",mask->bY );
        }
        
        p += sprintf(p, "),");
    }
    
    if (!isMask || mask->tX != defaultIndex || mask->tY != defaultIndex || mask->tZ != defaultIndex) {
        if (blockSpec) {
            p += sprintf(p, "(");
        }
        else {
            p += sprintf(p, "<<<(");
        }
        
        if (!isMask || mask->tX != defaultIndex) { 
            p += sprintf(p, "%d",mask->tX ); 
        }
        
        if (mask->tY != defaultIndex) { 
            p += sprintf(p, ",%d",mask->tY );
        } else
        if (mask->tZ != defaultIndex) { 
            p += sprintf(p, ",");
        }
        
        if (mask->tZ != defaultIndex) { 
            p += sprintf(p, ",%d",mask->tZ );
        }
        
        p += sprintf(p, ")>>>");
    }
    
    *p= 0;
    
    return result;
}

    static Bool isWhiteSpace(char c, Bool newLine) {
        return (c==' ') || (c=='\t') || (newLine && (c=='\n'));
    } 

    static Bool pulloff( String *input, uInt *result, Bool last )
    {
        uInt    n;
        Char   *e;
        String  s= *input;

        if (!s) { return True; }

        n= strtol(s,&e,0); 

        if (s != e) {
            s= e; *result= n;
        }

        while (isWhiteSpace(*s,False)) { s++; }

        if (*s) {
            if (last || *s != ',') { return False; }
            s++;
        }

       *input= s;
        return True;
    }

Bool ConstructGpuCoords(Bool isMask, String blockRepr, String threadRepr, gpudbgGpuCoordinates current, gpudbgGpuCoordinates *mask)
{
    struct gpudbgGpuCoordinates result;
    Int    defaultIndex= isMask ? -1 : 0;

    if (current) {
        if (!blockRepr) {
            /* We're not updating block coords, so keep them
             * to what they currently are */
            result.bX = current->bX;
            result.bY = current->bY;
        }
        else {
            /* We have current block coords, but we're updating them
             * so set them to the default before processing */
            result.bX = defaultIndex;
            result.bY = defaultIndex;
        }

        if (!threadRepr) {
            /* We're not updating thread coords, so keep them
             * to what they currently are */
            result.tX = current->tX;
            result.tY = current->tY;
            result.tZ = current->tZ;
        }
        else {
            /* We have current thread coords, but we're updating them
             * so set them to the default before processing */
            result.tX = defaultIndex;
            result.tY = defaultIndex;
            result.tZ = defaultIndex;
        }
    }
    else {
        /* We have no current coordinates, set all to default */
        result.bX = defaultIndex;
        result.bY = defaultIndex;
        result.tX = defaultIndex;
        result.tY = defaultIndex;
        result.tZ = defaultIndex;
    }

    if (!pulloff( &blockRepr,  &result.bX, False )) { return False; }
    if (!pulloff( &blockRepr,  &result.bY, True  )) { return False; }
    if (!pulloff( &threadRepr, &result.tX, False )) { return False; }
    if (!pulloff( &threadRepr, &result.tY, False )) { return False; }
    if (!pulloff( &threadRepr, &result.tZ, True  )) { return False; }

    if ( isMask
      && result.bX == -1
      && result.bY == -1
      && result.tX == -1
      && result.tY == -1
      && result.tZ == -1
       ) {    
       *mask= NULL;
    } else { 
        *mask  = xmalloc(sizeof result);
        memcpy(*mask, &result, sizeof(result));
    }

    return True;
}

/*
 * gpu block and thread commands:
 */

int parse_gpu_thread( char **ptok, char **pgpu_block_mask_repr, char** pgpu_thread_mask_repr, int *pFoundEntry);

static void switch_to_xheap( gpudbgGpuCoordinates *c)
{
    if (*c) {
        gpudbgGpuCoordinates r= xmalloc(sizeof(*r));
       *r= **c;
        gpudbg->DeleteGpuCoords(*c);
       *c= r;
    }
}

static void parse_thread_block(int is_mask, char *arg, gpudbgGpuCoordinates *coords) 
{
  char *gpu_block_mask_repr           = NULL;
  char *gpu_thread_mask_repr          = NULL;
  int   foundEntry                    =    0;
  int   tid                           = TIDGET(inferior_ptid);
  gpudbgGpuCoordinates current_thread = gpudbg->GetCurrentGpuCoords( tid );


  while (arg && *arg) {
     if (parse_gpu_thread(&arg, &gpu_block_mask_repr, &gpu_thread_mask_repr, &foundEntry)) {
         while (*arg == ' ' || *arg == '\t') { arg++; }
     } else {
         error (_("Junk at end of arguments."));
     }
  }
  
  if (!ConstructGpuCoords(is_mask,gpu_block_mask_repr,gpu_thread_mask_repr,
              current_thread, coords)) {
      error (_("Error in CUDA block/thread specification."));
  }
  
  switch_to_xheap(coords);
}

static void
switch_to_gpu_thread (gpudbgGpuCoordinates thread, int full_massage)
{
  int                  tid            = TIDGET(inferior_ptid);
  gpudbgGpuCoordinates current_thread = gpudbg->GetCurrentGpuCoords( tid );
  Bool                 isCurrent      = gpudbg->ThreadEqual(thread,current_thread);
  
  switch_to_xheap(&current_thread);
  
  if (!isCurrent) {
      if (gpudbg->SetCurrentGpuCoords(tid,thread)) {
          flush_cached_frames ();
          registers_changed ();
          stop_pc = read_pc ();
          gpudbg_massage_call_chain(inferior_ptid,full_massage);
          select_frame (get_current_frame ());
      } else {
          error (_("CUDA thread not known or not active"));
      }
  }
}

static unsigned
read_gpu_pc (gpudbgGpuCoordinates thread)
{
  int                  tid            = TIDGET(inferior_ptid);
  gpudbgGpuCoordinates currentThread  = gpudbg->GetCurrentGpuCoords(tid);
  unsigned tmp_pc;
  
  if (gpudbg->SetCurrentGpuCoords(tid, thread)) {
      tmp_pc = read_pc ();
      gpudbg->SetCurrentGpuCoords(tid, currentThread);
      return tmp_pc;
  } else {
      error (_("CUDA thread not known or not active"));
  }
}

void
cuda_thread_select (char *tidstr)
{
  char *gpu_block_mask_repr   = NULL;
  char *gpu_thread_mask_repr = NULL;
  char prefixed_tidstr[1000];
  gpudbgGpuCoordinates gpu_thread;
  Bool switched;
  
  gpudbgGpuCoordinates current_thread= gpudbg->GetCurrentGpuCoords( TIDGET(inferior_ptid) );
  switch_to_xheap(&current_thread);
        
  if (!current_thread) {
      error (_("No CUDA threads."));
  }
  
  if (!tidstr)
    {
      /* Don't generate an error, just say which thread is current. */
      if (target_has_stack) {
          printf_filtered (_("Current CUDA Thread %s "), 
                  GetGpuCoordsRepresentation(True, current_thread));
          print_stack_frame (get_selected_frame (NULL), 0, LOCATION);
      } else {
          error (_("No stack."));
      }
      return;
    }

  parse_thread_block(False,tidstr,&gpu_thread);

  switch_to_gpu_thread (gpu_thread, 1);

  gpudbg->DeleteGpuCoords(previous_gpu_thread);
  previous_gpu_thread= gpu_thread;
  
  printf_filtered (_("Switching to %s "),
                     GetGpuCoordsRepresentation(True, gpu_thread)
                  );

  print_stack_frame (get_selected_frame (NULL), 0, SRC_AND_LOC);
}

    typedef struct {
        gpudbgGpuCoordinates current;
        gpudbgGpuCoordinates mask;
        int                  tid;
    } ThreadInfoRec;

    static Bool gpuThreadInfo(gpudbgTBNumbers n, gpudbgGpuCoordinates thread, ThreadInfoRec *rec)
    {
      if (!rec->mask || gpudbg->ThreadMatch(rec->mask,thread)) {
        switch_to_gpu_thread (thread,0);

        if (gpudbg->ThreadEqual(thread,rec->current))
          printf_filtered (_("* "));
        else
          printf_filtered (_("  "));

        printf_filtered (_("%s "), GetGpuCoordsRepresentation(True, thread));
        print_stack_frame (get_selected_frame (NULL), 0, LOCATION);
      }
      
      return False;
    }

static unsigned summaryThreadPC = 0;
static struct gpudbgGpuCoordinates sSummaryCoords = {0};
static struct gpudbgGpuCoordinates cSummaryCoords = {0};
static struct gpudbgGpuCoordinates pSummaryCoords = {0};
static Bool summaryInitialized = False;

static Bool gpuThreadSummaryInfo(gpudbgTBNumbers n, 
                                 gpudbgGpuCoordinates thread,
                                 ThreadInfoRec *rec)
{
    if (!rec->mask || gpudbg->ThreadMatch(rec->mask,thread)) {
        
        /* Set previous and current coordinates */
        pSummaryCoords = cSummaryCoords;
        cSummaryCoords = *thread;
        
        if (!summaryInitialized) {
            summaryThreadPC = read_gpu_pc(&cSummaryCoords);
            sSummaryCoords = cSummaryCoords;
            summaryInitialized = True;
        }
        
        /* Check for a different PC */
        if (summaryThreadPC != read_gpu_pc(&cSummaryCoords)) {
            
            /* Gather range start gpu thread context */
            switch_to_gpu_thread (&sSummaryCoords, 0);
            
            /* Print range */
            printf_filtered (_("%s ... "), 
                    GetGpuCoordsRepresentation(True, &sSummaryCoords));
            printf_filtered (_("%s "), 
                    GetGpuCoordsRepresentation(True, &pSummaryCoords));
            print_stack_frame (get_selected_frame (NULL), 0, LOCATION);
            
            /* Reset range start pc */
            summaryThreadPC = read_gpu_pc(&cSummaryCoords);
            
            /* Reset range start coordinates */
            sSummaryCoords = cSummaryCoords;
        }
        
    }
    
    return False;
}

static int
do_captured_thread_info (struct ui_out *uiout, void *recp)
{
  ThreadInfoRec *rec= recp;

  gpudbg->ThreadsTraverse(rec->tid, (gpudbgThreadFun)gpuThreadInfo, rec);
  return GDB_RC_OK;
}

static int
do_captured_thread_summary (struct ui_out *uiout, void *recp)
{
    ThreadInfoRec *rec = recp;

    /* Setup state */
    sSummaryCoords = (struct gpudbgGpuCoordinates){0};
    cSummaryCoords = (struct gpudbgGpuCoordinates){0};
    pSummaryCoords = (struct gpudbgGpuCoordinates){0};
    summaryInitialized = False;
    summaryThreadPC = 0;

    /* Iterate over all threads */
    gpudbg->ThreadsTraverse(rec->tid, 
            (gpudbgThreadFun)gpuThreadSummaryInfo, rec);

    /* If we never initialized, there were no threads */
    if (!summaryInitialized) {
        printf_filtered (_("No active CUDA threads!"));
    }
    else {
        switch_to_gpu_thread (&cSummaryCoords, 0);
        printf_filtered (_("%s ... "), 
                GetGpuCoordsRepresentation(True, &sSummaryCoords));
        printf_filtered (_("%s "), 
                GetGpuCoordsRepresentation(True, &cSummaryCoords));
        print_stack_frame (get_selected_frame (NULL), 0, LOCATION);
    }

    return GDB_RC_OK;
}

static void
info_gpu_threads_command_aux (char *arg, int from_tty, int summarize)
{
  int                  tid            = TIDGET(inferior_ptid);
  gpudbgGpuCoordinates current_thread = gpudbg->GetCurrentGpuCoords(tid);
  gpudbgGpuCoordinates printmask = NULL;

  struct frame_id saved_frame_id;
  struct frame_info *cur_frame;
  ThreadInfoRec rec;

  switch_to_xheap(&current_thread);

  if (!current_thread) {
      error (_("No active GPU function."));
  }
  
  saved_frame_id = get_frame_id (get_selected_frame (NULL));
 
  /* 

     Unnecessary here...
   
     parse_thread_block(True,arg,&printmask);

  */
       
  rec.tid     = tid;      
  rec.mask    = printmask;
  rec.current = current_thread;

  if (summarize)
      catch_exceptions_with_msg (uiout, do_captured_thread_summary, &rec,
              NULL, RETURN_MASK_ALL);
  else
      catch_exceptions_with_msg (uiout, do_captured_thread_info, &rec,
              NULL, RETURN_MASK_ALL);

  switch_to_gpu_thread(current_thread,1);

  cur_frame = frame_find_by_id (saved_frame_id);
  
  if (cur_frame == NULL)
    {
      /* Ooops, can't restore, tell user where we are.  */
      warning (_("Couldn't restore frame in current thread, at frame 0"));
      print_stack_frame (get_selected_frame (NULL), 0, LOCATION);
    }
  else
    {
      select_frame (cur_frame);
      /* re-show current frame. */
      show_stack_frame (cur_frame);
    }
}

static void
info_gpu_threads_all_command (char *arg, int from_tty)
{
    info_gpu_threads_command_aux (arg, from_tty, 0);
}

static void
info_gpu_threads_command (char *arg, int from_tty)
{
    info_gpu_threads_command_aux (arg, from_tty, 1);
}

static void
info_gpu_state_command (char *arg, int from_tty)
{
    char *output= gpudbg->GetGPUState(arg);
    
    printf_filtered (_("%s"),output);
    
    gpudbg->DeleteGPUState(output);
}


static void
info_gpu_memory_state_command (char *arg, int from_tty)
{
    char *output= gpudbg->GetMemoryState(arg);
    
    printf_filtered (_("%s"),output);
    
    gpudbg->DeleteMemoryState(output);
}


    typedef struct {
        char *cmd;
        char *saved_cmd;
        int   from_tty;
    } ThreadExecRec;

    static Bool gpuThreadExec(gpudbgTBNumbers n, gpudbgGpuCoordinates thread, ThreadExecRec *rec)
    {
        switch_to_gpu_thread (thread,0);
        printf_filtered (_("%s "), GetGpuCoordsRepresentation(True, thread));
        
	execute_command (rec->cmd, rec->from_tty);
	strcpy (rec->cmd, rec->saved_cmd);	/* Restore exact command used previously */
      
        return False;
    }

    
static void
info_cuda_command (char *arg, int from_tty)
{
    char *s;

    /* info cuda */
    if (!arg)
        error (_("Must specify one of the following:  threads, state"));

    /* info cuda threads */
    if (!strncmp(arg, "threads", strlen("threads"))) {
        s = arg + strlen("threads");
        while (isWhiteSpace(*s, False)) { 
            s++;
        }
        if (*s) {
            /* info cuda threads all */
            if (!strcmp(s, "all")) {
                info_gpu_threads_all_command(NULL, from_tty);
            }
            else {
                printf_filtered (_("Unrecognized option:  %s\n"), arg);
                error(_("Must specify 'info cuda threads all'.\n"));
            }
        }
        else {
            /* info cuda threads */
            info_gpu_threads_command(NULL, from_tty);
        }
    }
    else if (!strcmp(arg, "state"))
        /* info gpu state */
        info_gpu_state_command(NULL, from_tty);
    else {
        /* Undefined command */
        printf_filtered (_("Unrecognized option: %s\n"), arg);
        error (_("Must specify one of the following:  threads, state"));
    }

#if 0
   if (!strcmp(arg, "memory")) {
        info_gpu_memory_state_command(NULL, from_tty);
   }
#endif

}

    
static void
thread_apply_all_command (char *cmd, int from_tty)
{
  struct cleanup *saved_cmd_cleanup_chain;
  ThreadExecRec rec;
  char *saved_cmd;

  if (cmd == NULL || *cmd == '\000') return; 
  
  /* Save a copy of the command in case it is clobbered by
     execute_command */
  rec.saved_cmd = xstrdup (cmd);
  rec.cmd       = cmd;
  rec.from_tty  = from_tty;

  saved_cmd_cleanup_chain = make_cleanup (xfree, (void *) rec.saved_cmd);
  
  gpudbg->ThreadsTraverse(TIDGET(inferior_ptid), (gpudbgThreadFun)gpuThreadExec, &rec);

  do_cleanups (saved_cmd_cleanup_chain);
}

/* ------------------------------------------------------------------------- */

/* Provide a prototype to silence -Wmissing-prototypes.  */
void _initialize_amd64_linux_nat (void);

void
_initialize_amd64_linux_nat (void)
{
  struct target_ops *t;

  amd64_native_gregset32_reg_offset = amd64_linux_gregset32_reg_offset;
  amd64_native_gregset32_num_regs = I386_LINUX_NUM_REGS;
  amd64_native_gregset64_reg_offset = amd64_linux_gregset64_reg_offset;
  amd64_native_gregset64_num_regs = AMD64_LINUX_NUM_REGS;

  gdb_assert (ARRAY_SIZE (amd64_linux_gregset32_reg_offset)
	      == amd64_native_gregset32_num_regs);
  gdb_assert (ARRAY_SIZE (amd64_linux_gregset64_reg_offset)
	      == amd64_native_gregset64_num_regs);

  /* Fill in the generic GNU/Linux methods.  */
  t = linux_target ();

  /* Override the GNU/Linux inferior startup hook.  */
  super_post_startup_inferior = t->to_post_startup_inferior;
  t->to_post_startup_inferior = amd64_linux_child_post_startup_inferior;

  /* Add our register access methods.  */
  t->to_fetch_registers = amd64_linux_fetch_inferior_registers;
  t->to_store_registers = amd64_linux_store_inferior_registers;

  /* Register the target.  */
  linux_nat_add_target (t);

  /* Override the default ptrace resume method.  */
  saved_to_kill               = t->to_kill;
  saved_to_mourn_inferior     = t->to_mourn_inferior;
  saved_to_wait               = t->to_wait;
  saved_to_resume             = t->to_resume;
  saved_to_create_inferior    = t->to_create_inferior;
  saved_to_xfer_partial       = t->to_xfer_partial;
  saved_to_fetch_registers    = t->to_fetch_registers;
  saved_to_store_registers    = t->to_store_registers;
  saved_to_insert_breakpoint  = t->to_insert_breakpoint;
  saved_to_remove_breakpoint  = t->to_remove_breakpoint;

  t->to_kill                  = wrapper_to_kill;
  t->to_mourn_inferior        = wrapper_to_mourn_inferior;
  t->to_resume                = wrapper_to_resume;
  t->to_wait                  = wrapper_to_wait;
  t->to_create_inferior       = wrapper_to_create_inferior;
  t->to_xfer_partial          = wrapper_to_xfer_partial;
  t->to_fetch_registers       = wrapper_to_fetch_registers;
  t->to_store_registers       = wrapper_to_store_registers;
  t->to_insert_breakpoint     = wrapper_to_insert_breakpoint;
  t->to_remove_breakpoint     = wrapper_to_remove_breakpoint;
  
  {
      static struct cmd_list_element *gpu_thread_cmd_list = NULL;
      static struct cmd_list_element *gpu_block_cmd_list  = NULL;

      add_info ("cuda", info_cuda_command,
	    _("[threads] [state] - Print CUDA information."));

#if 0
      add_cmd ("applyall", class_run, thread_apply_all_command,
	   _("Apply a command to all CUDA threads."), &gpu_thread_cmd_list);

      add_cmd ("applyall", class_run, block_apply_all_command,
	   _("Apply a command to all CUDA blocks."), &gpu_block_cmd_list);
#endif
  }
}
