/* Native-dependent code for GNU/Linux i386.

   Copyright (C) 1999, 2000, 2001, 2002, 2003, 2004, 2005, 2006
   Free Software Foundation, Inc.

   This file is part of GDB.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor,
   Boston, MA 02110-1301, USA.  */

#include "defs.h"
#include "inferior.h"
#include "gdbcore.h"
#include "regcache.h"
#include "target.h"
#include "linux-nat.h"

#include "gdb_assert.h"
#include "gdb_string.h"
#include <sys/ptrace.h>
#include <sys/user.h>
#include <sys/procfs.h>

#ifdef HAVE_SYS_REG_H
#include <sys/reg.h>
#endif

#ifndef ORIG_EAX
#define ORIG_EAX -1
#endif

#ifdef HAVE_SYS_DEBUGREG_H
#include <sys/debugreg.h>
#endif

#ifndef DR_FIRSTADDR
#define DR_FIRSTADDR 0
#endif

#ifndef DR_LASTADDR
#define DR_LASTADDR 3
#endif

#ifndef DR_STATUS
#define DR_STATUS 6
#endif

#ifndef DR_CONTROL
#define DR_CONTROL 7
#endif

/* Prototypes for supply_gregset etc.  */
#include "gregset.h"

#include "i387-tdep.h"
#include "i386-tdep.h"
#include "i386-linux-tdep.h"

/* Defines ps_err_e, struct ps_prochandle.  */
#include "gdb_proc_service.h"


/* The register sets used in GNU/Linux ELF core-dumps are identical to
   the register sets in `struct user' that is used for a.out
   core-dumps, and is also used by `ptrace'.  The corresponding types
   are `elf_gregset_t' for the general-purpose registers (with
   `elf_greg_t' the type of a single GP register) and `elf_fpregset_t'
   for the floating-point registers.

   Those types used to be available under the names `gregset_t' and
   `fpregset_t' too, and this file used those names in the past.  But
   those names are now used for the register sets used in the
   `mcontext_t' type, and have a different size and layout.  */

/* Mapping between the general-purpose registers in `struct user'
   format and GDB's register array layout.  */
static int regmap[] = 
{
  EAX, ECX, EDX, EBX,
  UESP, EBP, ESI, EDI,
  EIP, EFL, CS, SS,
  DS, ES, FS, GS,
  -1, -1, -1, -1,		/* st0, st1, st2, st3 */
  -1, -1, -1, -1,		/* st4, st5, st6, st7 */
  -1, -1, -1, -1,		/* fctrl, fstat, ftag, fiseg */
  -1, -1, -1, -1,		/* fioff, foseg, fooff, fop */
  -1, -1, -1, -1,		/* xmm0, xmm1, xmm2, xmm3 */
  -1, -1, -1, -1,		/* xmm4, xmm5, xmm6, xmm6 */
  -1,				/* mxcsr */
  ORIG_EAX
};

/* Which ptrace request retrieves which registers?
   These apply to the corresponding SET requests as well.  */

#define GETREGS_SUPPLIES(regno) \
  ((0 <= (regno) && (regno) <= 15) || (regno) == I386_LINUX_ORIG_EAX_REGNUM)

#define GETFPXREGS_SUPPLIES(regno) \
  (I386_ST0_REGNUM <= (regno) && (regno) < I386_SSE_NUM_REGS)

/* Does the current host support the GETREGS request?  */
int have_ptrace_getregs =
#ifdef HAVE_PTRACE_GETREGS
  1
#else
  0
#endif
;

/* Does the current host support the GETFPXREGS request?  The header
   file may or may not define it, and even if it is defined, the
   kernel will return EIO if it's running on a pre-SSE processor.

   My instinct is to attach this to some architecture- or
   target-specific data structure, but really, a particular GDB
   process can only run on top of one kernel at a time.  So it's okay
   for this to be a simple variable.  */
int have_ptrace_getfpxregs =
#ifdef HAVE_PTRACE_GETFPXREGS
  1
#else
  0
#endif
;


/* Support for the user struct.  */

/* Return the address of register REGNUM.  BLOCKEND is the value of
   u.u_ar0, which should point to the registers.  */

CORE_ADDR
register_u_addr (CORE_ADDR blockend, int regnum)
{
  return (blockend + 4 * regmap[regnum]);
}

/* Return the size of the user struct.  */

int
kernel_u_size (void)
{
  return (sizeof (struct user));
}


/* Accessing registers through the U area, one at a time.  */

/* Fetch one register.  */

static void
fetch_register (int regno)
{
  int tid;
  int val;

  gdb_assert (!have_ptrace_getregs);
  if (cannot_fetch_register (regno))
    {
      regcache_raw_supply (current_regcache, regno, NULL);
      return;
    }

  /* GNU/Linux LWP ID's are process ID's.  */
  tid = TIDGET (inferior_ptid);
  if (tid == 0)
    tid = PIDGET (inferior_ptid); /* Not a threaded program.  */

  errno = 0;
  val = ptrace (PTRACE_PEEKUSER, tid, register_addr (regno, 0), 0);
  if (errno != 0)
    error (_("Couldn't read register %s (#%d): %s."), REGISTER_NAME (regno),
	   regno, safe_strerror (errno));

  regcache_raw_supply (current_regcache, regno, &val);
}

/* Store one register. */

static void
store_register (int regno)
{
  int tid;
  int val;

  gdb_assert (!have_ptrace_getregs);
  if (cannot_store_register (regno))
    return;

  /* GNU/Linux LWP ID's are process ID's.  */
  tid = TIDGET (inferior_ptid);
  if (tid == 0)
    tid = PIDGET (inferior_ptid); /* Not a threaded program.  */

  errno = 0;
  regcache_raw_collect (current_regcache, regno, &val);
  ptrace (PTRACE_POKEUSER, tid, register_addr (regno, 0), val);
  if (errno != 0)
    error (_("Couldn't write register %s (#%d): %s."), REGISTER_NAME (regno),
	   regno, safe_strerror (errno));
}


/* Transfering the general-purpose registers between GDB, inferiors
   and core files.  */

/* Fill GDB's register array with the general-purpose register values
   in *GREGSETP.  */

void
supply_gregset (elf_gregset_t *gregsetp)
{
  elf_greg_t *regp = (elf_greg_t *) gregsetp;
  int i;

  for (i = 0; i < I386_NUM_GREGS; i++)
    regcache_raw_supply (current_regcache, i, regp + regmap[i]);

  if (I386_LINUX_ORIG_EAX_REGNUM < NUM_REGS)
    regcache_raw_supply (current_regcache, I386_LINUX_ORIG_EAX_REGNUM,
			 regp + ORIG_EAX);
}

/* Fill register REGNO (if it is a general-purpose register) in
   *GREGSETPS with the value in GDB's register array.  If REGNO is -1,
   do this for all registers.  */

void
fill_gregset (elf_gregset_t *gregsetp, int regno)
{
  elf_greg_t *regp = (elf_greg_t *) gregsetp;
  int i;

  for (i = 0; i < I386_NUM_GREGS; i++)
    if (regno == -1 || regno == i)
      regcache_raw_collect (current_regcache, i, regp + regmap[i]);

  if ((regno == -1 || regno == I386_LINUX_ORIG_EAX_REGNUM)
      && I386_LINUX_ORIG_EAX_REGNUM < NUM_REGS)
    regcache_raw_collect (current_regcache, I386_LINUX_ORIG_EAX_REGNUM,
			  regp + ORIG_EAX);
}

#ifdef HAVE_PTRACE_GETREGS

/* Fetch all general-purpose registers from process/thread TID and
   store their values in GDB's register array.  */

static void
fetch_regs (int tid)
{
  elf_gregset_t regs;

  if (ptrace (PTRACE_GETREGS, tid, 0, (int) &regs) < 0)
    {
      if (errno == EIO)
	{
	  /* The kernel we're running on doesn't support the GETREGS
             request.  Reset `have_ptrace_getregs'.  */
	  have_ptrace_getregs = 0;
	  return;
	}

      perror_with_name (_("Couldn't get registers"));
    }

  supply_gregset (&regs);
}

/* Store all valid general-purpose registers in GDB's register array
   into the process/thread specified by TID.  */

static void
store_regs (int tid, int regno)
{
  elf_gregset_t regs;

  if (ptrace (PTRACE_GETREGS, tid, 0, (int) &regs) < 0)
    perror_with_name (_("Couldn't get registers"));

  fill_gregset (&regs, regno);
  
  if (ptrace (PTRACE_SETREGS, tid, 0, (int) &regs) < 0)
    perror_with_name (_("Couldn't write registers"));
}

#else

static void fetch_regs (int tid) {}
static void store_regs (int tid, int regno) {}

#endif


/* Transfering floating-point registers between GDB, inferiors and cores.  */

/* Fill GDB's register array with the floating-point register values in
   *FPREGSETP.  */

void 
supply_fpregset (elf_fpregset_t *fpregsetp)
{
  i387_supply_fsave (current_regcache, -1, fpregsetp);
}

/* Fill register REGNO (if it is a floating-point register) in
   *FPREGSETP with the value in GDB's register array.  If REGNO is -1,
   do this for all registers.  */

void
fill_fpregset (elf_fpregset_t *fpregsetp, int regno)
{
  i387_fill_fsave ((char *) fpregsetp, regno);
}

#ifdef HAVE_PTRACE_GETREGS

/* Fetch all floating-point registers from process/thread TID and store
   thier values in GDB's register array.  */

static void
fetch_fpregs (int tid)
{
  elf_fpregset_t fpregs;

  if (ptrace (PTRACE_GETFPREGS, tid, 0, (int) &fpregs) < 0)
    perror_with_name (_("Couldn't get floating point status"));

  supply_fpregset (&fpregs);
}

/* Store all valid floating-point registers in GDB's register array
   into the process/thread specified by TID.  */

static void
store_fpregs (int tid, int regno)
{
  elf_fpregset_t fpregs;

  if (ptrace (PTRACE_GETFPREGS, tid, 0, (int) &fpregs) < 0)
    perror_with_name (_("Couldn't get floating point status"));

  fill_fpregset (&fpregs, regno);

  if (ptrace (PTRACE_SETFPREGS, tid, 0, (int) &fpregs) < 0)
    perror_with_name (_("Couldn't write floating point status"));
}

#else

static void fetch_fpregs (int tid) {}
static void store_fpregs (int tid, int regno) {}

#endif


/* Transfering floating-point and SSE registers to and from GDB.  */

#ifdef HAVE_PTRACE_GETFPXREGS

/* Fill GDB's register array with the floating-point and SSE register
   values in *FPXREGSETP.  */

void
supply_fpxregset (elf_fpxregset_t *fpxregsetp)
{
  i387_supply_fxsave (current_regcache, -1, fpxregsetp);
}

/* Fill register REGNO (if it is a floating-point or SSE register) in
   *FPXREGSETP with the value in GDB's register array.  If REGNO is
   -1, do this for all registers.  */

void
fill_fpxregset (elf_fpxregset_t *fpxregsetp, int regno)
{
  i387_fill_fxsave ((char *) fpxregsetp, regno);
}

/* Fetch all registers covered by the PTRACE_GETFPXREGS request from
   process/thread TID and store their values in GDB's register array.
   Return non-zero if successful, zero otherwise.  */

static int
fetch_fpxregs (int tid)
{
  elf_fpxregset_t fpxregs;

  if (! have_ptrace_getfpxregs)
    return 0;

  if (ptrace (PTRACE_GETFPXREGS, tid, 0, (int) &fpxregs) < 0)
    {
      if (errno == EIO)
	{
	  have_ptrace_getfpxregs = 0;
	  return 0;
	}

      perror_with_name (_("Couldn't read floating-point and SSE registers"));
    }

  supply_fpxregset (&fpxregs);
  return 1;
}

/* Store all valid registers in GDB's register array covered by the
   PTRACE_SETFPXREGS request into the process/thread specified by TID.
   Return non-zero if successful, zero otherwise.  */

static int
store_fpxregs (int tid, int regno)
{
  elf_fpxregset_t fpxregs;

  if (! have_ptrace_getfpxregs)
    return 0;
  
  if (ptrace (PTRACE_GETFPXREGS, tid, 0, &fpxregs) == -1)
    {
      if (errno == EIO)
	{
	  have_ptrace_getfpxregs = 0;
	  return 0;
	}

      perror_with_name (_("Couldn't read floating-point and SSE registers"));
    }

  fill_fpxregset (&fpxregs, regno);

  if (ptrace (PTRACE_SETFPXREGS, tid, 0, &fpxregs) == -1)
    perror_with_name (_("Couldn't write floating-point and SSE registers"));

  return 1;
}

#else

static int fetch_fpxregs (int tid) { return 0; }
static int store_fpxregs (int tid, int regno) { return 0; }

#endif /* HAVE_PTRACE_GETFPXREGS */


/* Transferring arbitrary registers between GDB and inferior.  */

/* Check if register REGNO in the child process is accessible.
   If we are accessing registers directly via the U area, only the
   general-purpose registers are available.
   All registers should be accessible if we have GETREGS support.  */
   
int
cannot_fetch_register (int regno)
{
  gdb_assert (regno >= 0 && regno < NUM_REGS);
  return (!have_ptrace_getregs && regmap[regno] == -1);
}

int
cannot_store_register (int regno)
{
  gdb_assert (regno >= 0 && regno < NUM_REGS);
  return (!have_ptrace_getregs && regmap[regno] == -1);
}

/* Fetch register REGNO from the child process.  If REGNO is -1, do
   this for all registers (including the floating point and SSE
   registers).  */

static void
i386_linux_fetch_inferior_registers (int regno)
{
  int tid;

  /* Use the old method of peeking around in `struct user' if the
     GETREGS request isn't available.  */
  if (!have_ptrace_getregs)
    {
      int i;

      for (i = 0; i < NUM_REGS; i++)
	if (regno == -1 || regno == i)
	  fetch_register (i);

      return;
    }

  /* GNU/Linux LWP ID's are process ID's.  */
  tid = TIDGET (inferior_ptid);
  if (tid == 0)
    tid = PIDGET (inferior_ptid); /* Not a threaded program.  */

  /* Use the PTRACE_GETFPXREGS request whenever possible, since it
     transfers more registers in one system call, and we'll cache the
     results.  But remember that fetch_fpxregs can fail, and return
     zero.  */
  if (regno == -1)
    {
      fetch_regs (tid);

      /* The call above might reset `have_ptrace_getregs'.  */
      if (!have_ptrace_getregs)
	{
	  i386_linux_fetch_inferior_registers (regno);
	  return;
	}

      if (fetch_fpxregs (tid))
	return;
      fetch_fpregs (tid);
      return;
    }

  if (GETREGS_SUPPLIES (regno))
    {
      fetch_regs (tid);
      return;
    }

  if (GETFPXREGS_SUPPLIES (regno))
    {
      if (fetch_fpxregs (tid))
	return;

      /* Either our processor or our kernel doesn't support the SSE
	 registers, so read the FP registers in the traditional way,
	 and fill the SSE registers with dummy values.  It would be
	 more graceful to handle differences in the register set using
	 gdbarch.  Until then, this will at least make things work
	 plausibly.  */
      fetch_fpregs (tid);
      return;
    }

  internal_error (__FILE__, __LINE__,
		  _("Got request for bad register number %d."), regno);
}

/* Store register REGNO back into the child process.  If REGNO is -1,
   do this for all registers (including the floating point and SSE
   registers).  */
static void
i386_linux_store_inferior_registers (int regno)
{
  int tid;

  /* Use the old method of poking around in `struct user' if the
     SETREGS request isn't available.  */
  if (!have_ptrace_getregs)
    {
      int i;

      for (i = 0; i < NUM_REGS; i++)
	if (regno == -1 || regno == i)
	  store_register (i);

      return;
    }

  /* GNU/Linux LWP ID's are process ID's.  */
  tid = TIDGET (inferior_ptid);
  if (tid == 0)
    tid = PIDGET (inferior_ptid); /* Not a threaded program.  */

  /* Use the PTRACE_SETFPXREGS requests whenever possible, since it
     transfers more registers in one system call.  But remember that
     store_fpxregs can fail, and return zero.  */
  if (regno == -1)
    {
      store_regs (tid, regno);
      if (store_fpxregs (tid, regno))
	return;
      store_fpregs (tid, regno);
      return;
    }

  if (GETREGS_SUPPLIES (regno))
    {
      store_regs (tid, regno);
      return;
    }

  if (GETFPXREGS_SUPPLIES (regno))
    {
      if (store_fpxregs (tid, regno))
	return;

      /* Either our processor or our kernel doesn't support the SSE
	 registers, so just write the FP registers in the traditional
	 way.  */
      store_fpregs (tid, regno);
      return;
    }

  internal_error (__FILE__, __LINE__,
		  _("Got request to store bad register number %d."), regno);
}


/* Support for debug registers.  */

static unsigned long
i386_linux_dr_get (int regnum)
{
  int tid;
  unsigned long value;

  /* FIXME: kettenis/2001-01-29: It's not clear what we should do with
     multi-threaded processes here.  For now, pretend there is just
     one thread.  */
  tid = PIDGET (inferior_ptid);

  /* FIXME: kettenis/2001-03-27: Calling perror_with_name if the
     ptrace call fails breaks debugging remote targets.  The correct
     way to fix this is to add the hardware breakpoint and watchpoint
     stuff to the target vector.  For now, just return zero if the
     ptrace call fails.  */
  errno = 0;
  value = ptrace (PTRACE_PEEKUSER, tid,
		  offsetof (struct user, u_debugreg[regnum]), 0);
  if (errno != 0)
#if 0
    perror_with_name (_("Couldn't read debug register"));
#else
    return 0;
#endif

  return value;
}

static void
i386_linux_dr_set (int regnum, unsigned long value)
{
  int tid;

  /* FIXME: kettenis/2001-01-29: It's not clear what we should do with
     multi-threaded processes here.  For now, pretend there is just
     one thread.  */
  tid = PIDGET (inferior_ptid);

  errno = 0;
  ptrace (PTRACE_POKEUSER, tid,
	  offsetof (struct user, u_debugreg[regnum]), value);
  if (errno != 0)
    perror_with_name (_("Couldn't write debug register"));
}

void
i386_linux_dr_set_control (unsigned long control)
{
  i386_linux_dr_set (DR_CONTROL, control);
}

void
i386_linux_dr_set_addr (int regnum, CORE_ADDR addr)
{
  gdb_assert (regnum >= 0 && regnum <= DR_LASTADDR - DR_FIRSTADDR);

  i386_linux_dr_set (DR_FIRSTADDR + regnum, addr);
}

void
i386_linux_dr_reset_addr (int regnum)
{
  gdb_assert (regnum >= 0 && regnum <= DR_LASTADDR - DR_FIRSTADDR);

  i386_linux_dr_set (DR_FIRSTADDR + regnum, 0L);
}

unsigned long
i386_linux_dr_get_status (void)
{
  return i386_linux_dr_get (DR_STATUS);
}


/* Called by libthread_db.  Returns a pointer to the thread local
   storage (or its descriptor).  */

ps_err_e
ps_get_thread_area (const struct ps_prochandle *ph, 
		    lwpid_t lwpid, int idx, void **base)
{
  /* NOTE: cagney/2003-08-26: The definition of this buffer is found
     in the kernel header <asm-i386/ldt.h>.  It, after padding, is 4 x
     4 byte integers in size: `entry_number', `base_addr', `limit',
     and a bunch of status bits.

     The values returned by this ptrace call should be part of the
     regcache buffer, and ps_get_thread_area should channel its
     request through the regcache.  That way remote targets could
     provide the value using the remote protocol and not this direct
     call.

     Is this function needed?  I'm guessing that the `base' is the
     address of a a descriptor that libthread_db uses to find the
     thread local address base that GDB needs.  Perhaps that
     descriptor is defined by the ABI.  Anyway, given that
     libthread_db calls this function without prompting (gdb
     requesting tls base) I guess it needs info in there anyway.  */
  unsigned int desc[4];
  gdb_assert (sizeof (int) == 4);

#ifndef PTRACE_GET_THREAD_AREA
#define PTRACE_GET_THREAD_AREA 25
#endif

  if (ptrace (PTRACE_GET_THREAD_AREA, lwpid,
	      (void *) idx, (unsigned long) &desc) < 0)
    return PS_ERR;

  *(int *)base = desc[1];
  return PS_OK;
}


/* The instruction for a GNU/Linux system call is:
       int $0x80
   or 0xcd 0x80.  */

static const unsigned char linux_syscall[] = { 0xcd, 0x80 };

#define LINUX_SYSCALL_LEN (sizeof linux_syscall)

/* The system call number is stored in the %eax register.  */
#define LINUX_SYSCALL_REGNUM I386_EAX_REGNUM

/* We are specifically interested in the sigreturn and rt_sigreturn
   system calls.  */

#ifndef SYS_sigreturn
#define SYS_sigreturn		0x77
#endif
#ifndef SYS_rt_sigreturn
#define SYS_rt_sigreturn	0xad
#endif

/* Offset to saved processor flags, from <asm/sigcontext.h>.  */
#define LINUX_SIGCONTEXT_EFLAGS_OFFSET (64)

/* Resume execution of the inferior process.
   If STEP is nonzero, single-step it.
   If SIGNAL is nonzero, give it that signal.  */

static void
i386_linux_resume (ptid_t ptid, int step, enum target_signal signal)
{
  int pid = PIDGET (ptid);

  int request = PTRACE_CONT;

  if (pid == -1)
    /* Resume all threads.  */
    /* I think this only gets used in the non-threaded case, where "resume
       all threads" and "resume inferior_ptid" are the same.  */
    pid = PIDGET (inferior_ptid);

  if (step)
    {
      CORE_ADDR pc = read_pc_pid (pid_to_ptid (pid));
      gdb_byte buf[LINUX_SYSCALL_LEN];

      request = PTRACE_SINGLESTEP;

      /* Returning from a signal trampoline is done by calling a
         special system call (sigreturn or rt_sigreturn, see
         i386-linux-tdep.c for more information).  This system call
         restores the registers that were saved when the signal was
         raised, including %eflags.  That means that single-stepping
         won't work.  Instead, we'll have to modify the signal context
         that's about to be restored, and set the trace flag there.  */

      /* First check if PC is at a system call.  */
      if (read_memory_nobpt (pc, buf, LINUX_SYSCALL_LEN) == 0
	  && memcmp (buf, linux_syscall, LINUX_SYSCALL_LEN) == 0)
	{
	  int syscall = read_register_pid (LINUX_SYSCALL_REGNUM,
	                                   pid_to_ptid (pid));

	  /* Then check the system call number.  */
	  if (syscall == SYS_sigreturn || syscall == SYS_rt_sigreturn)
	    {
	      CORE_ADDR sp = read_register (I386_ESP_REGNUM);
	      CORE_ADDR addr = sp;
	      unsigned long int eflags;

	      if (syscall == SYS_rt_sigreturn)
		addr = read_memory_integer (sp + 8, 4) + 20;

	      /* Set the trace flag in the context that's about to be
                 restored.  */
	      addr += LINUX_SIGCONTEXT_EFLAGS_OFFSET;
	      read_memory (addr, (gdb_byte *) &eflags, 4);
	      eflags |= 0x0100;
	      write_memory (addr, (gdb_byte *) &eflags, 4);
	    }
	}
    }

  if (ptrace (request, pid, 0, target_signal_to_host (signal)) == -1)
    perror_with_name (("ptrace"));
}

static void (*super_post_startup_inferior) (ptid_t ptid);

static void
i386_linux_child_post_startup_inferior (ptid_t ptid)
{
  i386_cleanup_dregs ();
  super_post_startup_inferior (ptid);
}

/* ------------------------------------------------------------------------- */

#include "nvGpuDebugServer.h"
#include "nvPtxStorage.h"
#include "stdRangeMap.h"

gpudbgGpuCoordinates previous_gpu_thread;

static stdRangeMap_t ptxAddresses;
static CORE_ADDR     saved_raw_pc;
static int           saved_raw_pc_valid;

static void    (*saved_to_kill) (void);
static void    (*saved_to_mourn_inferior) (void);
static ptid_t  (*saved_to_wait) (ptid_t, struct target_waitstatus *);
static void    (*saved_to_resume) (ptid_t tid, int i, enum target_signal ts);
static void    (*saved_to_create_inferior) (char *, char *, char **, int);
static void    (*saved_to_fetch_registers) (int regno);
static void    (*saved_to_store_registers) (int regno);
static int     (*saved_to_insert_breakpoint) (struct bp_target_info *bp_tgt);
static int     (*saved_to_remove_breakpoint) (struct bp_target_info *bp_tgt);

static LONGEST (*saved_to_xfer_partial) (struct target_ops *ops,
				enum target_object object, const char *annex,
				gdb_byte *readbuf, const gdb_byte *writebuf,
				ULONGEST offset, LONGEST len);


static void wrapper_to_kill(void)
{
    gpudbg_cleanup();
    saved_to_kill();
}


void gdbg_flush_saved_thread_state()
{
    if (saved_raw_pc_valid) { 
        regcache_raw_write(current_regcache,PC_REGNUM,&saved_raw_pc);
        saved_raw_pc_valid= 0;
    }
}


static void wrapper_to_mourn_inferior(void)
{
    if (ptxAddresses) {
        gpudbg->rangemapDelete(ptxAddresses);
        ptxAddresses= NULL;
    }
    
    saved_raw_pc_valid= 0;
    gpudbg_cleanup();
    saved_to_mourn_inferior();
}


static void wrapper_to_resume (ptid_t tid, int sstep, enum target_signal ts)
{
    gpudbgGPUContext gpuContext;
    int gpuContextTid;

    gdbg_flush_saved_thread_state();

    gpuContextTid = tid.pid == -1 
        ? inferior_ptid.lwp ? inferior_ptid.lwp : inferior_ptid.pid
        : tid.lwp ? tid.lwp : tid.pid;
    
    /*
     * See patch at infrun.c, function currently_stepping (Return Code is 2):
     */
    if (sstep != 2) {
       /*
        * From gnu-nat.c, line 1924:
        *       TID  STEP:
        *       -1   true   Single step the current thread allowing other threads to run.
        *       -1   false  Continue the current thread allowing other threads to run.
        *       X    true   Single step the given thread, don't allow any others to run.
        *       X    false  Continue the given thread, do not allow any others to run.
        *       (Where X, of course, is anything except -1)
        */

        gpudbg->ResumeGpus(gpuContextTid, sstep, tid.pid == -1 ? True : False);
    }

    /* Now we determine how to resume the host.  We always need the host to run in
     * the case of a GPU context, otherwise we cannot detect GPU termination. */
    if (gpudbg->GetGPUContext(gpuContextTid, &gpuContext)) {
        /* We have a context, so let the host go, we'll send the trap for the GPU */
        saved_to_resume(tid,0,ts);

        /* If we single stepped the GPU, send the trap here.
         * The target layer handles this for breakpoints. */
        if (sstep) {
            kill(gpuContextTid, SIGTRAP);
        }
    }
    else {
        /* Always resume the host if there is no gpu context */
        saved_to_resume(tid,sstep,ts);
    }

}

static ptid_t wrapper_to_wait (ptid_t tid, struct target_waitstatus *ws)
{
    gpudbgGPUContext gpuContext;
    ptid_t r= saved_to_wait(tid,ws);

    /* We need to make sure that saved_to_wait hasn't neglected the GPU.
     * It is possible that the lwp field inside of r (above) has been changed
     * to a different lwp in the host process. We check to see that we had a
     * gpu thread in our previous context, then we also check to see that we
     * still have an active GPU context. If so, we set the return value of
     * this function accordingly. */

    /* Check to see if we have a gpu context with the current r.lwp.
     * We only update if we do not have a gpu context */
    if (ptid_get_lwp(r) != 0 && !gpudbg->GetGPUContext(r.lwp, &gpuContext)){
        if (previous_gpu_thread && ptid_get_lwp(inferior_ptid) != 0){
            /* We know that we had a previous gpu thread and that there is
             * a specified lwp field in inferior_ptid */
            if (gpudbg->GetGPUContext(inferior_ptid.lwp, &gpuContext)){
                /* We know that the lwp field matches that registered by the client */
                r = inferior_ptid;
            }
        }
    }

   /*
    * Most cleanly, gpudbg_initialize() should have been called
    * at inferior creation. It is done lazily here because initialization
    * attaches to special debugger symbols in the cuda library, and this 
    * library is loaded later during executable startup. We therefore
    * simply wait until we see it appear. 
    * It is guaranteed that we eventually catch it, because thread creation 
    * and dll loading will trigger hidden debugger breakpoints.
    */
    gpudbg_initialize();
    gpudbg->SuspendGpus();
    
   /*
    * The clean way of getting to know all host addresses that 
    * correspond with device function code is to place an observer
    * in the shared library module, to get notified whenever a new
    * shared library gets loaded or unloaded. However, this is another
    * intrusion in the gdb code, so we do it lazily: we rescan the 
    * list of currently loaded libraries each time we get control back:
    */
    if (ptxAddresses) {
        gpudbg->rangemapDelete(ptxAddresses);
        ptxAddresses= NULL;
    }

    return r;
}

static void wrapper_to_create_inferior (char *a, char *b, char **c, int d)
{
    saved_to_create_inferior(a,b,c,d);
    gpudbg_initialize();
}

static LONGEST wrapper_to_xfer_partial(struct target_ops *ops,
				enum target_object object, const char *annex,
				gdb_byte *readbuf, const gdb_byte *writebuf,
				ULONGEST offset, LONGEST len)
{
    LONGEST result= 0;
    
    if (object == TARGET_OBJECT_MEMORY) {
        if (readbuf) {
            result= gpudbg->ReadMemory(offset,readbuf,len);
        } else
        if (writebuf) {
            result= gpudbg->WriteMemory(offset,writebuf,len);
        }
    }
    
    if (!result) {
        result= saved_to_xfer_partial(ops,object,annex,readbuf,writebuf,offset,len);
    }
            
    return result;
}

static void wrapper_to_fetch_registers (int regno)
{
    saved_to_fetch_registers (regno);
}

static void wrapper_to_store_registers (int regno)
{
    saved_to_store_registers (regno);
}

static int wrapper_to_insert_breakpoint (struct bp_target_info *bp_tgt)
{
  if (gpudbg->SetBreakpoint(bp_tgt->placed_address,bp_tgt->owner->owner->gpu_thread,1)) {
    return 0;
  } else {
    return saved_to_insert_breakpoint(bp_tgt);
  }
}

static int wrapper_to_remove_breakpoint (struct bp_target_info *bp_tgt)
{
  if (gpudbg->SetBreakpoint(bp_tgt->placed_address,bp_tgt->owner->owner->gpu_thread,0)) {
    return 0;
  } else {
    return saved_to_remove_breakpoint(bp_tgt);
  }
}


/* ------------------------------------------------------------------------- */

/*
 * The full_massage parameter having value False causes gpudbg_massage_call_chain 
 * to only patch the topmost stack frame (describing the current pc).
 * This is used when it is known that no stack frame up/down traversal
 * is necessary (see info gpu threads command).
 * A better solution would be to let full massaging be called lazily by
 * function get_prev_frame:
 */

void gpudbg_massage_call_chain( ptid_t ptid, int full_massage )
{
    gpudbgGPUContext gpuContext;
    int tid = TIDGET (ptid);

   /*
    * Inspect the break pc, and if this is a GPU address,
    * then precache the frame_info chain up to the device code
    * 'caller', while kicking out all frames that correspond to the cuda
    * runtime launching functions. This makes it appear in the debugger
    * as if the application directly 'called' the device function.
    */
    if (gpudbg->GetGPUContext(tid, &gpuContext)) { 
        struct frame_info *bottom,*frame;
        int                morePrev = 1;
        CORE_ADDR          low,high,raw_pc;
        char              *name;
        
       /*
        * Restore original pc in host thread, otherwise
        * frame unwinding will use the wrong debugger info:
        */
        bottom = get_current_frame();
        
        gdbg_flush_saved_thread_state();

        raw_pc = read_register_pid (PC_REGNUM, ptid);
        kickout_intermediate_frames(bottom,bottom,ADDR_BITS_REMOVE(raw_pc));
        
       /*
        * While unwinding the stack, search for the 
        * device function stub called by the application,
        * and then kick out all intermediate frames from the
        * frame cache:
        */
        frame = get_prev_frame(bottom);
        
        if ( full_massage
          && find_pc_partial_function (gpuContext.functionWrapper, &name, &low, &high)
           ) {
            while (frame) {
                CORE_ADDR pc = get_frame_pc(frame);
                if (pc >= low && pc < high) { break; }
                frame = get_prev_frame(frame);
            }

            while (frame && morePrev) {
                frame = get_prev_frame(frame);
                morePrev--;
            }
        }

        if (frame) {
            CORE_ADDR pc_val   = gpuContext.pc;
            saved_raw_pc       = raw_pc;
            saved_raw_pc_valid = 1;
            
            get_frame_id(frame); // evaluate the frame descriptor
            kickout_intermediate_frames(frame,bottom,gpuContext.pc);
            regcache_raw_write(current_regcache,PC_REGNUM,&pc_val);
        }
    }
}

/* ------------------------------------------------------------------------- */

#include "filenames.h"
#include "objfiles.h"

/*
 * Read all known source link sections (which consist of pairs of filenames,
 * <host source file,device source file>. If pst->filename appears to be such
 * a host source filename, then also load the symbtab of the corresponding 
 * device source file name. ALthough duplicates are not likely, neither host 
 * source file name nor device source file name is supposed to be unique 
 * (they may be temp file names). If they are not, then we might at most 
 * load too many symbol tables, but this is not a problem for the debugger.
 * (this function is a patch to dwarf2read.c)
 */
void gpudbg_loadDeviceInfo( char *fileName, struct partial_symtab *pst)
{
    struct section_table *secp;
    for (secp = current_target.to_sections;
         secp < current_target.to_sections_end;
         secp++) {
        if (strcmp(secp->the_bfd_section->name,PTX_SOURCELINK_SEGMENT_NAME)==0) {
            int   sectionSize = secp->endaddr - secp->addr; 
            char *sourceLinks = alloca(sectionSize);
            char *sourceBound = sourceLinks + sectionSize;

            target_read_memory(secp->addr,sourceLinks,sectionSize);

            while (sourceLinks < sourceBound) {
                int   size     = strlen(sourceLinks)+1;
                char *hostName = sourceLinks;
                char *devName  = strchr(sourceLinks,',');

                // --- we need some sanity checking here ---

                *(devName++) = 0;
                sourceLinks += size;

                if (FILENAME_CMP (hostName, fileName) == 0) {
                    struct partial_symtab *ps;
                    ALL_OBJFILE_PSYMTABS (pst->objfile, ps)
                    {
                        if (FILENAME_CMP (devName, ps->filename) == 0) {
                            if (!ps->readin)
                                dwarf2_psymtab_to_symtab(ps);
                        }
                    }
                }
                
               /*
                * creep past section padding:
                */
                while ((sourceLinks < sourceBound) && *sourceLinks==0) {
                    sourceLinks++;
                }
            }
        }
    }
}

/* ------------------------------------------------------------------------- */

#include "filenames.h"
#include "objfiles.h"

/*
 * Read all known ptx virtual code sections, and record 
 * the virtual ptx code addresses in a range map:
 */
int gpudbg_is_ptx_address( CORE_ADDR address )
{
    if (!ptxAddresses) {
        struct section_table *secp;
        
        ptxAddresses = gpudbg->rangemapCreate();
        
        for (secp = current_target.to_sections;
             secp < current_target.to_sections_end;
             secp++) {
            if (strcmp(secp->the_bfd_section->name,PTX_NVSEGMENT_NAME(10))==0) {
                rangemapDomain_t dom;
                
                dom.start  = secp->addr;
                dom.length = secp->endaddr - secp->addr; 
                
                gpudbg->rangemapDefine( ptxAddresses,  dom, (Pointer)(Address)True );
            }
        }
    }
    
    return gpudbg->rangemapApply( ptxAddresses, address ) != NULL;
}

/* ------------------------------------------------------------------------- */

#include "ui-out.h"
#include "command.h"
#include "gdb.h"
#include "exceptions.h"
#include "gdbcmd.h"

String GetGpuCoordsRepresentation(Bool isMask, gpudbgGpuCoordinates mask)
{
    Int    defaultIndex= isMask ? -1 : 0;
    Int    blockSpec = 0;

    static Char  result[1000];
           Char *p= result;
    
    if (!isMask || mask->bX != defaultIndex || mask->bY != defaultIndex) {
        blockSpec = 1;

        p += sprintf(p, "<<<(");
        
        if (!isMask || mask->bX != defaultIndex) { 
            p += sprintf(p, "%d",mask->bX ); 
        }
        
        if (mask->bY != defaultIndex) { 
            p += sprintf(p, ",%d",mask->bY );
        }
        
        p += sprintf(p, "),");
    }
    
    if (!isMask || mask->tX != defaultIndex || mask->tY != defaultIndex || mask->tZ != defaultIndex) {
        if (blockSpec) {
            p += sprintf(p, "(");
        }
        else {
            p += sprintf(p, "<<<(");
        }
        
        if (!isMask || mask->tX != defaultIndex) { 
            p += sprintf(p, "%d",mask->tX ); 
        }
        
        if (mask->tY != defaultIndex) { 
            p += sprintf(p, ",%d",mask->tY );
        } else
        if (mask->tZ != defaultIndex) { 
            p += sprintf(p, ",");
        }
        
        if (mask->tZ != defaultIndex) { 
            p += sprintf(p, ",%d",mask->tZ );
        }
        
        p += sprintf(p, ")>>>");
    }
    
    *p= 0;
    
    return result;
}

    static Bool isWhiteSpace(char c, Bool newLine) {
        return (c==' ') || (c=='\t') || (newLine && (c=='\n'));
    } 

    static Bool pulloff( String *input, uInt *result, Bool last )
    {
        uInt    n;
        Char   *e;
        String  s= *input;

        if (!s) { return True; }

        n= strtol(s,&e,0); 

        if (s != e) {
            s= e; *result= n;
        }

        while (isWhiteSpace(*s,False)) { s++; }

        if (*s) {
            if (last || *s != ',') { return False; }
            s++;
        }

       *input= s;
        return True;
    }

Bool ConstructGpuCoords(Bool isMask, String blockRepr, String threadRepr, gpudbgGpuCoordinates current, gpudbgGpuCoordinates *mask)
{
    struct gpudbgGpuCoordinates result;
    Int    defaultIndex= isMask ? -1 : 0;

    if (current) {
        if (!blockRepr) {
            /* We're not updating block coords, so keep them
             * to what they currently are */
            result.bX = current->bX;
            result.bY = current->bY;
        }
        else {
            /* We have current block coords, but we're updating them
             * so set them to the default before processing */
            result.bX = defaultIndex;
            result.bY = defaultIndex;
        }

        if (!threadRepr) {
            /* We're not updating thread coords, so keep them
             * to what they currently are */
            result.tX = current->tX;
            result.tY = current->tY;
            result.tZ = current->tZ;
        }
        else {
            /* We have current thread coords, but we're updating them
             * so set them to the default before processing */
            result.tX = defaultIndex;
            result.tY = defaultIndex;
            result.tZ = defaultIndex;
        }
    }
    else {
        /* We have no current coordinates, set all to default */
        result.bX = defaultIndex;
        result.bY = defaultIndex;
        result.tX = defaultIndex;
        result.tY = defaultIndex;
        result.tZ = defaultIndex;
    }

    if (!pulloff( &blockRepr,  &result.bX, False )) { return False; }
    if (!pulloff( &blockRepr,  &result.bY, True  )) { return False; }
    if (!pulloff( &threadRepr, &result.tX, False )) { return False; }
    if (!pulloff( &threadRepr, &result.tY, False )) { return False; }
    if (!pulloff( &threadRepr, &result.tZ, True  )) { return False; }

    if ( isMask
      && result.bX == -1
      && result.bY == -1
      && result.tX == -1
      && result.tY == -1
      && result.tZ == -1
       ) {    
       *mask= NULL;
    } else { 
        *mask  = xmalloc(sizeof result);
        memcpy(*mask, &result, sizeof(result));
    }

    return True;
}

/*
 * gpu block and thread commands:
 */

int parse_gpu_thread( char **ptok, char **pgpu_block_mask_repr, char** pgpu_thread_mask_repr, int *pFoundEntry);

static void switch_to_xheap( gpudbgGpuCoordinates *c)
{
    if (*c) {
        gpudbgGpuCoordinates r= xmalloc(sizeof(*r));
       *r= **c;
        gpudbg->DeleteGpuCoords(*c);
       *c= r;
    }
}

static void parse_thread_block(int is_mask, char *arg, gpudbgGpuCoordinates *coords) 
{
  char *gpu_block_mask_repr           = NULL;
  char *gpu_thread_mask_repr          = NULL;
  int   foundEntry                    =    0;
  int   tid                           = TIDGET(inferior_ptid);
  gpudbgGpuCoordinates current_thread = gpudbg->GetCurrentGpuCoords( tid );


  while (arg && *arg) {
     if (parse_gpu_thread(&arg, &gpu_block_mask_repr, &gpu_thread_mask_repr, &foundEntry)) {
         while (*arg == ' ' || *arg == '\t') { arg++; }
     } else {
         error (_("Junk at end of arguments."));
     }
  }
  
  if (!ConstructGpuCoords(is_mask,gpu_block_mask_repr,gpu_thread_mask_repr,
              current_thread, coords)) {
      error (_("Error in CUDA block/thread specification."));
  }
  
  switch_to_xheap(coords);
}

static void
switch_to_gpu_thread (gpudbgGpuCoordinates thread, int full_massage)
{
  int                  tid            = TIDGET(inferior_ptid);
  gpudbgGpuCoordinates current_thread = gpudbg->GetCurrentGpuCoords( tid );
  Bool                 isCurrent      = gpudbg->ThreadEqual(thread,current_thread);
  
  switch_to_xheap(&current_thread);
  
  if (!isCurrent) {
      if (gpudbg->SetCurrentGpuCoords(tid,thread)) {
          flush_cached_frames ();
          registers_changed ();
          stop_pc = read_pc ();
          gpudbg_massage_call_chain(inferior_ptid,full_massage);
          select_frame (get_current_frame ());
      } else {
          error (_("CUDA thread not known or not active"));
      }
  }
}

static unsigned
read_gpu_pc (gpudbgGpuCoordinates thread)
{
  int                  tid            = TIDGET(inferior_ptid);
  gpudbgGpuCoordinates currentThread  = gpudbg->GetCurrentGpuCoords(tid);
  unsigned tmp_pc;
  
  if (gpudbg->SetCurrentGpuCoords(tid, thread)) {
      tmp_pc = read_pc ();
      gpudbg->SetCurrentGpuCoords(tid, currentThread);
      return tmp_pc;
  } else {
      error (_("CUDA thread not known or not active"));
  }
}

void
cuda_thread_select (char *tidstr)
{
  char *gpu_block_mask_repr   = NULL;
  char *gpu_thread_mask_repr = NULL;
  char prefixed_tidstr[1000];
  gpudbgGpuCoordinates gpu_thread;
  Bool switched;
  
  gpudbgGpuCoordinates current_thread= gpudbg->GetCurrentGpuCoords( TIDGET(inferior_ptid) );
  switch_to_xheap(&current_thread);
        
  if (!current_thread) {
      error (_("No CUDA threads."));
  }
  
  if (!tidstr)
    {
      /* Don't generate an error, just say which thread is current. */
      if (target_has_stack) {
          printf_filtered (_("Current CUDA Thread %s "), 
                  GetGpuCoordsRepresentation(True, current_thread));
          print_stack_frame (get_selected_frame (NULL), 0, LOCATION);
      } else {
          error (_("No stack."));
      }
      return;
    }

  parse_thread_block(False,tidstr,&gpu_thread);

  switch_to_gpu_thread (gpu_thread, 1);

  gpudbg->DeleteGpuCoords(previous_gpu_thread);
  previous_gpu_thread= gpu_thread;
  
  printf_filtered (_("Switching to %s "),
                     GetGpuCoordsRepresentation(True, gpu_thread)
                  );

  print_stack_frame (get_selected_frame (NULL), 0, SRC_AND_LOC);
}

    typedef struct {
        gpudbgGpuCoordinates current;
        gpudbgGpuCoordinates mask;
        int                  tid;
    } ThreadInfoRec;

    static Bool gpuThreadInfo(gpudbgTBNumbers n, gpudbgGpuCoordinates thread, ThreadInfoRec *rec)
    {
      if (!rec->mask || gpudbg->ThreadMatch(rec->mask,thread)) {
        switch_to_gpu_thread (thread,0);

        if (gpudbg->ThreadEqual(thread,rec->current))
          printf_filtered (_("* "));
        else
          printf_filtered (_("  "));

        printf_filtered (_("%s "), GetGpuCoordsRepresentation(True, thread));
        print_stack_frame (get_selected_frame (NULL), 0, LOCATION);
      }
      
      return False;
    }

static unsigned summaryThreadPC = 0;
static struct gpudbgGpuCoordinates sSummaryCoords = {0};
static struct gpudbgGpuCoordinates cSummaryCoords = {0};
static struct gpudbgGpuCoordinates pSummaryCoords = {0};
static Bool summaryInitialized = False;

static Bool gpuThreadSummaryInfo(gpudbgTBNumbers n, 
                                 gpudbgGpuCoordinates thread,
                                 ThreadInfoRec *rec)
{
    if (!rec->mask || gpudbg->ThreadMatch(rec->mask,thread)) {
        
        /* Set previous and current coordinates */
        pSummaryCoords = cSummaryCoords;
        cSummaryCoords = *thread;
        
        if (!summaryInitialized) {
            summaryThreadPC = read_gpu_pc(&cSummaryCoords);
            sSummaryCoords = cSummaryCoords;
            summaryInitialized = True;
        }
        
        /* Check for a different PC */
        if (summaryThreadPC != read_gpu_pc(&cSummaryCoords)) {
            
            /* Gather range start gpu thread context */
            switch_to_gpu_thread (&sSummaryCoords, 0);
            
            /* Print range */
            printf_filtered (_("%s ... "), 
                    GetGpuCoordsRepresentation(True, &sSummaryCoords));
            printf_filtered (_("%s "), 
                    GetGpuCoordsRepresentation(True, &pSummaryCoords));
            print_stack_frame (get_selected_frame (NULL), 0, LOCATION);
            
            /* Reset range start pc */
            summaryThreadPC = read_gpu_pc(&cSummaryCoords);
            
            /* Reset range start coordinates */
            sSummaryCoords = cSummaryCoords;
        }
        
    }
    
    return False;
}

static int
do_captured_thread_info (struct ui_out *uiout, void *recp)
{
  ThreadInfoRec *rec= recp;

  gpudbg->ThreadsTraverse(rec->tid, (gpudbgThreadFun)gpuThreadInfo, rec);
  return GDB_RC_OK;
}

static int
do_captured_thread_summary (struct ui_out *uiout, void *recp)
{
    ThreadInfoRec *rec = recp;

    /* Setup state */
    sSummaryCoords = (struct gpudbgGpuCoordinates){0};
    cSummaryCoords = (struct gpudbgGpuCoordinates){0};
    pSummaryCoords = (struct gpudbgGpuCoordinates){0};
    summaryInitialized = False;
    summaryThreadPC = 0;

    /* Iterate over all threads */
    gpudbg->ThreadsTraverse(rec->tid, 
            (gpudbgThreadFun)gpuThreadSummaryInfo, rec);

    /* If we never initialized, there were no threads */
    if (!summaryInitialized) {
        printf_filtered (_("No active CUDA threads!"));
    }
    else {
        switch_to_gpu_thread (&cSummaryCoords, 0);
        printf_filtered (_("%s ... "), 
                GetGpuCoordsRepresentation(True, &sSummaryCoords));
        printf_filtered (_("%s "), 
                GetGpuCoordsRepresentation(True, &cSummaryCoords));
        print_stack_frame (get_selected_frame (NULL), 0, LOCATION);
    }

    return GDB_RC_OK;
}

static void
info_gpu_threads_command_aux (char *arg, int from_tty, int summarize)
{
  int                  tid            = TIDGET(inferior_ptid);
  gpudbgGpuCoordinates current_thread = gpudbg->GetCurrentGpuCoords(tid);
  gpudbgGpuCoordinates printmask = NULL;

  struct frame_id saved_frame_id;
  struct frame_info *cur_frame;
  ThreadInfoRec rec;

  switch_to_xheap(&current_thread);

  if (!current_thread) {
      error (_("No active GPU function."));
  }
  
  saved_frame_id = get_frame_id (get_selected_frame (NULL));
 
  /* 

     Unnecessary here...
   
     parse_thread_block(True,arg,&printmask);

  */
       
  rec.tid     = tid;      
  rec.mask    = printmask;
  rec.current = current_thread;

  if (summarize)
      catch_exceptions_with_msg (uiout, do_captured_thread_summary, &rec,
              NULL, RETURN_MASK_ALL);
  else
      catch_exceptions_with_msg (uiout, do_captured_thread_info, &rec,
              NULL, RETURN_MASK_ALL);

  switch_to_gpu_thread(current_thread,1);

  cur_frame = frame_find_by_id (saved_frame_id);
  
  if (cur_frame == NULL)
    {
      /* Ooops, can't restore, tell user where we are.  */
      warning (_("Couldn't restore frame in current thread, at frame 0"));
      print_stack_frame (get_selected_frame (NULL), 0, LOCATION);
    }
  else
    {
      select_frame (cur_frame);
      /* re-show current frame. */
      show_stack_frame (cur_frame);
    }
}

static void
info_gpu_threads_all_command (char *arg, int from_tty)
{
    info_gpu_threads_command_aux (arg, from_tty, 0);
}

static void
info_gpu_threads_command (char *arg, int from_tty)
{
    info_gpu_threads_command_aux (arg, from_tty, 1);
}

static void
info_gpu_state_command (char *arg, int from_tty)
{
    char *output= gpudbg->GetGPUState(arg);
    
    printf_filtered (_("%s"),output);
    
    gpudbg->DeleteGPUState(output);
}


static void
info_gpu_memory_state_command (char *arg, int from_tty)
{
    char *output= gpudbg->GetMemoryState(arg);
    
    printf_filtered (_("%s"),output);
    
    gpudbg->DeleteMemoryState(output);
}


    typedef struct {
        char *cmd;
        char *saved_cmd;
        int   from_tty;
    } ThreadExecRec;

    static Bool gpuThreadExec(gpudbgTBNumbers n, gpudbgGpuCoordinates thread, ThreadExecRec *rec)
    {
        switch_to_gpu_thread (thread,0);
        printf_filtered (_("%s "), GetGpuCoordsRepresentation(True, thread));
        
	execute_command (rec->cmd, rec->from_tty);
	strcpy (rec->cmd, rec->saved_cmd);	/* Restore exact command used previously */
      
        return False;
    }

    
static void
info_cuda_command (char *arg, int from_tty)
{
    char *s;

    /* info cuda */
    if (!arg)
        error (_("Must specify one of the following:  threads, state"));

    /* info cuda threads */
    if (!strncmp(arg, "threads", strlen("threads"))) {
        s = arg + strlen("threads");
        while (isWhiteSpace(*s, False)) { 
            s++;
        }
        if (*s) {
            /* info cuda threads all */
            if (!strcmp(s, "all")) {
                info_gpu_threads_all_command(NULL, from_tty);
            }
            else {
                printf_filtered (_("Unrecognized option:  %s\n"), arg);
                error(_("Must specify 'info cuda threads all'.\n"));
            }
        }
        else {
            /* info cuda threads */
            info_gpu_threads_command(NULL, from_tty);
        }
    }
    else if (!strcmp(arg, "state"))
        /* info gpu state */
        info_gpu_state_command(NULL, from_tty);
    else {
        /* Undefined command */
        printf_filtered (_("Unrecognized option: %s\n"), arg);
        error (_("Must specify one of the following:  threads, state"));
    }

#if 0
   if (!strcmp(arg, "memory")) {
        info_gpu_memory_state_command(NULL, from_tty);
   }
#endif

}

    
static void
thread_apply_all_command (char *cmd, int from_tty)
{
  struct cleanup *saved_cmd_cleanup_chain;
  ThreadExecRec rec;
  char *saved_cmd;

  if (cmd == NULL || *cmd == '\000') return; 
  
  /* Save a copy of the command in case it is clobbered by
     execute_command */
  rec.saved_cmd = xstrdup (cmd);
  rec.cmd       = cmd;
  rec.from_tty  = from_tty;

  saved_cmd_cleanup_chain = make_cleanup (xfree, (void *) rec.saved_cmd);
  
  gpudbg->ThreadsTraverse(TIDGET(inferior_ptid), (gpudbgThreadFun)gpuThreadExec, &rec);

  do_cleanups (saved_cmd_cleanup_chain);
}

/* ------------------------------------------------------------------------- */


void
_initialize_i386_linux_nat (void)
{
  struct target_ops *t;

  /* Fill in the generic GNU/Linux methods.  */
  t = linux_target ();

  /* Override the default ptrace resume method.  */
  t->to_resume = i386_linux_resume;

  /* Override the GNU/Linux inferior startup hook.  */
  super_post_startup_inferior = t->to_post_startup_inferior;
  t->to_post_startup_inferior = i386_linux_child_post_startup_inferior;

  /* Add our register access methods.  */
  t->to_fetch_registers = i386_linux_fetch_inferior_registers;
  t->to_store_registers = i386_linux_store_inferior_registers;

  /* Register the target.  */
  linux_nat_add_target (t);
  
  /* Override the default ptrace resume method.  */
  saved_to_kill               = t->to_kill;
  saved_to_mourn_inferior     = t->to_mourn_inferior;
  saved_to_wait               = t->to_wait;
  saved_to_resume             = t->to_resume;
  saved_to_create_inferior    = t->to_create_inferior;
  saved_to_xfer_partial       = t->to_xfer_partial;
  saved_to_fetch_registers    = t->to_fetch_registers;
  saved_to_store_registers    = t->to_store_registers;
  saved_to_insert_breakpoint  = t->to_insert_breakpoint;
  saved_to_remove_breakpoint  = t->to_remove_breakpoint;

  t->to_kill                  = wrapper_to_kill;
  t->to_mourn_inferior        = wrapper_to_mourn_inferior;
  t->to_resume                = wrapper_to_resume;
  t->to_wait                  = wrapper_to_wait;
  t->to_create_inferior       = wrapper_to_create_inferior;
  t->to_xfer_partial          = wrapper_to_xfer_partial;
  t->to_fetch_registers       = wrapper_to_fetch_registers;
  t->to_store_registers       = wrapper_to_store_registers;
  t->to_insert_breakpoint     = wrapper_to_insert_breakpoint;
  t->to_remove_breakpoint     = wrapper_to_remove_breakpoint;
  
  {
      static struct cmd_list_element *gpu_thread_cmd_list = NULL;
      static struct cmd_list_element *gpu_block_cmd_list  = NULL;

      add_info ("cuda", info_cuda_command,
	    _("[threads] [state] - Print CUDA information."));

#if 0
      add_cmd ("applyall", class_run, thread_apply_all_command,
	   _("Apply a command to all CUDA threads."), &gpu_thread_cmd_list);

      add_cmd ("applyall", class_run, block_apply_all_command,
	   _("Apply a command to all CUDA blocks."), &gpu_block_cmd_list);
#endif
  }
}
